import os
from dotenv import load_dotenv
from typing import Any, Optional
from dataclasses import dataclass

load_dotenv()


@dataclass
class DtaProxyConfig:
    url: str
    key: str


class DtaProxy:

    _config: DtaProxyConfig = None

    @classmethod
    def set_keys(cls, *, key: str, url: Optional[str] = None):
        if not cls._config:
            cls._config = DtaProxyConfig(
                url=url,
                key=key,
            )
            return

        if url:
            cls._config.url = url

        if key:
            cls._config.key = key

    @classmethod
    def get_keys(cls, *, config: Optional[DtaProxyConfig] = None) -> DtaProxyConfig:
        return (
            DtaProxyConfig(
                url=config.url if config and config.url else cls._config.url,
                key=(config.key if config and config.key else cls._config.key),
            )
            if config
            else cls._config
        )

    @classmethod
    def url(cls) -> str:
        return cls._config.url

    @classmethod
    def key(cls) -> str:
        value = cls._config.key

        if not value:
            raise Exception("Environment var DTA_PROXY_KEY not found")

        return value

    @classmethod
    def openai_client(
        cls, *, openai: Optional[Any] = None, config: Optional[DtaProxyConfig] = None
    ):
        if not openai:
            import openai

        use_config = cls.get_keys(config=config)

        return openai.OpenAI(
            api_key=use_config.key,
            base_url=use_config.url,
        )


DTA_PROXY_URL = "https://proxy.dta.totvs.ai"

DtaProxy.set_keys(
    url=os.environ.get("DTA_PROXY_URL", DTA_PROXY_URL),
    key=os.environ.get("DTA_PROXY_KEY"),
)



import os
from typing import List, Optional


class DtaChatService:

    @classmethod
    def serve_chat(
        cls,
        api_config: Optional[dict] = None,
        sample_questions: Optional[List[str]] = None,
    ) -> str:
        html = cls.read_chat()

        html = html.replace(
            "/* API_CONFIG */",
            f"{api_config or cls.chat_api_config()};",
        )

        if sample_questions and len(sample_questions):
            html = html.replace(
                "/* SAMPLE_QUESTIONS */",
                f"{sample_questions};",
            )

        return html

    @classmethod
    def chat_api_config(cls) -> dict:
        baseUrl = "."

        str = """ {
            "apiInit": {
                "url": "{baseUrl}/welcome",
                "headers": {
                    "X-User-Auth": "localbypass",
                },
            },
            "apiMessage": {
                "url": "{baseUrl}/completion",
                "method": "POST",
            },
            "apiAudioTranscript": {
                "url": "{baseUrl}/audio",
                "method": "POST",
            },
            "apiFeedback": {
                "url": "{baseUrl}/feedback",
                "method": "POST",
            },
        }"""

        return str.replace("{baseUrl}", baseUrl)

    @classmethod
    def read_chat(cls):
        file_path = "./assets/dta_chat.html"

        docs_path = os.path.join(os.path.dirname(__file__), file_path)

        with open(docs_path, "r") as file:
            html = file.read()

        return html



from .services.dta_chat_service import DtaChatService
from .services.dta_observer_service import DtaObserver, UnsupportedAppError
from .services.dta_tracer_service import DtaLog, DtaTracer, DtaLogConfig
from .services.dta_proxy_service import DtaProxy, DtaProxyConfig



import copy
import traceback
from typing import Any, Optional
from .dta_tracer_service import DtaLog, DtaTracer
from dataclasses import dataclass


## Custom exceptions ##
class DtaCustomError(Exception):
    """Base DTA Logging exceptions"""

    def __init__(self, message: str = "DTA Logging error"):
        self.message = message
        super().__init__(self.message)


class UnsupportedAppError(DtaCustomError):
    """Raised when no an unsupported handler is used to observing"""

    def __init__(self):
        self.message = "Invalid handler: You must use the Goblet app context"
        super().__init__(self.message)


@dataclass
class DtaObserverAppConfig:
    header_trace_id: Optional[str]
    header_session_id: Optional[str]
    input: Optional[bool]
    output: Optional[bool]
    intercept_exceptions: Optional[bool]


@dataclass
class DtaAppContext:
    payload: dict[str, Any]
    request: Any


class DtaObserverApp:
    def register_app(self, app) -> None:
        # Identify which app is managing requests
        self._app = app
        self._load_app_context()

    def _load_app_context(self) -> DtaAppContext:
        if self._app:
            context = DtaAppContext(
                request=None,
                payload={},
            )
            try:
                # Trying Goblet
                from goblet import Goblet

                if isinstance(self._app, Goblet):
                    context.request = self._app.current_request
                    if hasattr(context.request, "method"):
                        # Retrieving Goblet request payload
                        if context.request.method in ["GET", "DELETE"]:
                            context.payload = context.request.args
                        elif context.request.method in ["PUT", "POST"]:
                            context.payload = context.request.get_json()

                    return context
            except Exception:
                pass

            # NOTE: extend for more APP Handlers here if it's needed (Flask, FastAPI, etc)

            raise UnsupportedAppError()

    def _intercepting(
        self,
        main,
        tracer: DtaTracer,
        observer_config: DtaObserverAppConfig,
        *args,
        **kwargs,
    ):
        context = self._load_app_context()

        if observer_config.input:
            self._process_input(
                tracer=tracer, observer_config=observer_config, context=context
            )

        try:
            main_result = main(*args, **kwargs, tracer=tracer)
        except Exception as e:
            if observer_config.intercept_exceptions:
                self._process_exceptions(
                    e, tracer=tracer, observer_config=observer_config, context=context
                )

            raise e

        if observer_config.output:
            self._process_output(
                response=copy.deepcopy(main_result),
                tracer=tracer,
                observer_config=observer_config,
                context=context,
            )

        return main_result

    def _process_input(
        self,
        tracer: DtaTracer,
        observer_config: DtaObserverAppConfig,
        context: DtaAppContext,
    ):
        trace_id = (
            context.request.headers.get(observer_config.header_trace_id)
            if observer_config.header_trace_id
            else None
        )
        session_id = (
            context.request.headers.get(observer_config.header_session_id)
            if observer_config.header_session_id
            else None
        )

        # TRACER input
        tracer.trace(
            id=trace_id,
            session_id=session_id,
            input=context.payload,
        )

    def _process_exceptions(
        self,
        ex: Exception,
        tracer: DtaTracer,
        observer_config: DtaObserverAppConfig,
        context: DtaAppContext,
    ):
        ex_detail = "".join(
            traceback.format_exception(ex, value=ex, tb=ex.__traceback__)
        )
        trace_id = (
            context.request.headers.get(observer_config.header_trace_id)
            if observer_config.header_trace_id
            else None
        )
        session_id = (
            context.request.headers.get(observer_config.header_session_id)
            if observer_config.header_session_id
            else None
        )

        # TRACER event
        tracer.event(
            trace_id=trace_id,
            session_id=session_id,
            name="exception",
            input=context.payload,
            output=ex_detail,
        )

    def _process_output(
        self,
        response,
        tracer: DtaTracer,
        observer_config: DtaObserverAppConfig,
        context: DtaAppContext,
    ):
        trace_id = (
            context.request.headers.get(observer_config.header_trace_id)
            if observer_config.header_trace_id
            else None
        )
        session_id = (
            context.request.headers.get(observer_config.header_session_id)
            if observer_config.header_session_id
            else None
        )

        # TRACER output
        err_code = (
            response.body["code"]
            if "code" in response.body and response.body["code"] >= 300
            else None
        )
        tracer.trace(
            id=trace_id,
            session_id=session_id,
            output=response.body,
            tags=["error", f"error-{err_code}"] if err_code else None,
        )


class DtaObserver(DtaObserverApp):
    _initial_tracer_values: dict[str, Any] = {}

    def __init__(
        self,
        app: any = None,
        **kwargs,
    ) -> None:
        self._initial_tracer_values = {
            key: value for key, value in kwargs.items() if key != "handler"
        }

        if app:
            self.register_app(app)

    def observe(
        self,
        header_trace_id="X-Dta-Thread",
        header_session_id="X-Dta-Session",
        input=True,
        output=True,
        intercept_exceptions=True,
    ):
        observer_config = DtaObserverAppConfig(
            header_trace_id=header_trace_id,
            header_session_id=header_session_id,
            input=input,
            output=output,
            intercept_exceptions=intercept_exceptions,
        )

        def _register_observer(func):
            def wrapper(*args, **kwargs):
                tracer = DtaTracer(**self._initial_tracer_values)

                return self._intercepting(
                    main=func,
                    tracer=tracer,
                    observer_config=observer_config,
                    *args,
                    **kwargs,
                )

            return wrapper

        return _register_observer

    def trace(self, **kwargs):
        merged_args = {
            **self._initial_tracer_values,
            **kwargs,
        }
        return DtaLog.trace(**merged_args)

    def event(self, **kwargs):
        return DtaLog.event(**kwargs)

    def score(
        self,
        *,
        name=str,
        value=float,
        trace_id=str,
        comment=str,
    ):
        return DtaLog.score(
            name=name,
            value=value,
            trace_id=trace_id,
            comment=comment,
        )



import os
from typing import Any, Literal, Optional, List
import datetime as dt
from langfuse import Langfuse
from dataclasses import dataclass
from dotenv import load_dotenv

load_dotenv()


@dataclass
class DtaLogConfig:
    url: str
    public_key: str
    secret_key: str


class DtaLog:
    """
    DtaTrace powered by Langfuse
    """

    _config: DtaLogConfig = None

    @classmethod
    def set_keys(cls, *, public_key: str, secret_key: str, url: Optional[str] = None):
        if not cls._config:
            cls._config = DtaLogConfig(
                url=url,
                public_key=public_key,
                secret_key=secret_key,
            )
            return

        if url:
            cls._config.url = url

        if public_key:
            cls._config.public_key = public_key

        if secret_key:
            cls._config.secret_key = secret_key

    @classmethod
    def get_keys(cls, *, config: Optional[DtaLogConfig] = None) -> DtaLogConfig:
        return (
            DtaLogConfig(
                url=config.url if config and config.url else cls._config.url,
                public_key=(
                    config.public_key
                    if config and config.public_key
                    else cls._config.public_key
                ),
                secret_key=(
                    config.secret_key
                    if config and config.secret_key
                    else cls._config.secret_key
                ),
            )
            if config
            else cls._config
        )

    @classmethod
    def _create(cls, *, config: Optional[DtaLogConfig] = None) -> Langfuse:
        use_config = cls.get_keys(config=config)

        return Langfuse(
            host=use_config.url,
            public_key=use_config.public_key,
            secret_key=use_config.secret_key,
        )

    @classmethod
    def score(
        cls,
        *,
        trace_id: str,
        name: str,
        value: float,
        comment: Optional[str] = None,
        config: Optional[DtaLogConfig] = None,
        **kwargs,
    ):
        if not trace_id:
            raise Exception("Score requires trace_id")

        return cls._create(config=config).score(
            **cls._props_not_none(
                name=name, value=value, trace_id=trace_id, comment=comment, **kwargs
            ),
        )

    @classmethod
    def trace(
        cls,
        *,
        config: Optional[DtaLogConfig] = None,
        id: Optional[str] = None,
        session_id: Optional[str] = None,
        name: Optional[str] = None,
        user_id: Optional[str] = None,
        version: Optional[str] = None,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        metadata: Optional[Any] = None,
        tags: Optional[List[str]] = None,
        **kwargs,
    ):
        return cls._create(config=config).trace(
            **cls._props_not_none(
                id=id,
                session_id=session_id,
                name=name,
                user_id=user_id,
                version=version,
                input=input,
                output=output,
                metadata=metadata,
                tags=cls._process_tags(tags),
                **kwargs,
            ),
        )

    @classmethod
    def event(
        cls,
        *,
        config: Optional[DtaLogConfig] = None,
        id: Optional[str] = None,
        trace_id: Optional[str] = None,
        name: Optional[str] = None,
        start_time: Optional[dt.datetime] = None,
        metadata: Optional[Any] = None,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        level: Optional[Literal["DEBUG", "DEFAULT", "WARNING", "ERROR"]] = None,
        status_message: Optional[str] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        return cls._create(config=config).event(
            **cls._props_not_none(
                id=id,
                trace_id=trace_id,
                name=name,
                start_time=start_time,
                metadata=metadata,
                input=input,
                output=output,
                level=level,
                status_message=status_message,
                version=version,
                **kwargs,
            ),
        )

    @classmethod
    def openai(
        cls,
        *,
        config: Optional[DtaLogConfig] = None,
    ):
        from langfuse.openai import openai as lf_openai

        use_config = cls.get_keys(config=config)

        lf_openai.langfuse_public_key = use_config.public_key
        lf_openai.langfuse_secret_key = use_config.secret_key
        lf_openai.langfuse_host = use_config.url

        return lf_openai

    @classmethod
    def openai_flush(cls, dta_openai: Any):
        if dta_openai and hasattr(dta_openai, "flush_langfuse"):
            dta_openai.flush_langfuse()

    @classmethod
    def langchain_handler(cls, config: Optional[DtaLogConfig] = None, **kwargs):
        from langfuse.callback import CallbackHandler

        use_config = cls.get_keys(config=config)

        return CallbackHandler(
            host=use_config.url,
            public_key=use_config.public_key,
            secret_key=use_config.secret_key,
            **kwargs,
        )

    @classmethod
    def _props_not_none(
        cls, *, init_props: dict[str, Any] = {}, **kwargs
    ) -> dict[str, Any]:
        return {
            **{key: value for key, value in init_props.items() if value is not None},
            **{key: value for key, value in kwargs.items() if value is not None},
        }

    @classmethod
    def _process_tags(
        cls,
        tags: List[str] = [],
        init_props: List[str] = [],
    ):
        return list(set((init_props or []) + (tags or []) + ["dta-trace"]))


class DtaTracer(DtaLog):
    _init_props: dict[str, Any]
    _tracer: Any = None
    _instance_config: DtaLogConfig = None

    def __init__(
        self,
        **kwargs,
    ) -> None:
        super().__init__()
        self._init_props = kwargs or {}

    def trace(
        self,
        *,
        id: Optional[str] = None,
        session_id: Optional[str] = None,
        name: Optional[str] = None,
        user_id: Optional[str] = None,
        version: Optional[str] = None,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        metadata: Optional[Any] = None,
        tags: Optional[List[str]] = None,
        **kwargs,
    ):
        if not self._tracer or (tags and len(tags)):
            self._init_props["tags"] = super()._process_tags(
                tags=tags,
                init_props=(
                    self._init_props["tags"] if "tags" in self._init_props else []
                ),
            )

        if self._tracer:
            return self._tracer.update(
                **super()._props_not_none(
                    init_props=self._init_props,
                    name=name,
                    user_id=user_id,
                    version=version,
                    input=input,
                    output=output,
                    metadata=metadata,
                    **kwargs,
                ),
            )

        self._tracer = super().trace(
            config=self._instance_config,
            **super()._props_not_none(
                init_props=self._init_props,
                id=id,
                session_id=session_id,
                name=name,
                user_id=user_id,
                version=version,
                input=input,
                output=output,
                metadata=metadata,
            ),
        )

        return self._tracer

    def event(
        self,
        *,
        id: Optional[str] = None,
        trace_id: Optional[str] = None,
        name: Optional[str] = None,
        start_time: Optional[dt.datetime] = None,
        metadata: Optional[Any] = None,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        level: Optional[Literal["DEBUG", "DEFAULT", "WARNING", "ERROR"]] = None,
        status_message: Optional[str] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        props = super()._props_not_none(
            init_props=self._init_props,
            id=id,
            trace_id=trace_id,
            name=name,
            start_time=start_time,
            metadata=metadata,
            input=input,
            output=output,
            level=level,
            status_message=status_message,
            version=version,
            **kwargs,
        )

        if not self._tracer:
            return super().event(config=self._instance_config, **props)

        return self._tracer.event(**props)

    def score(
        self,
        *,
        name: str,
        value: float,
        trace_id: Optional[str] = None,
        comment: Optional[str] = None,
        **kwargs,
    ):
        props = super()._props_not_none(
            init_props=self._init_props,
            name=name,
            value=value,
            trace_id=trace_id,
            comment=comment,
            **kwargs,
        )
        props.pop("tags", None)

        if not self._tracer:
            return super().score(config=self._instance_config, **props)

        return self._tracer.score(**props)

    def langchain_handler(self, **kwargs):
        """
        Usage example:

        from langchain_openai import ChatOpenAI

        def get_llm (tracer: DtaTracer) -> ChatOpenAI:
            handler = tracer.thread_langchain_handler()

            model = ChatOpenAI(
                temperature=0,
                model=model,
                callbacks=[handler],
            )

            return model
        """

        if not self._tracer:
            raise Exception("Tracer not created")

        return super().langchain_handler(
            config=self._instance_config,
            stateful_client=self._tracer,
            **kwargs,
        )

    def set_keys(self, *, public_key: str, secret_key: str, url: Optional[str] = None):
        if not self._instance_config:
            self._instance_config = DtaLogConfig(
                url=url,
                public_key=public_key,
                secret_key=secret_key,
            )
            return

        if url:
            self._instance_config.url = url

        if public_key:
            self._instance_config.public_key = public_key

        if secret_key:
            self._instance_config.secret_key = secret_key


DTA_LOGS_URL = "https://logs.dta.totvs.ai"

DtaLog.set_keys(
    url=os.environ.get("DTA_LOGS_URL", DTA_LOGS_URL),
    public_key=os.environ.get("DTA_LOGS_PUBLIC_KEY"),
    secret_key=os.environ.get("DTA_LOGS_SECRET_KEY"),
)



from setuptools import setup

readme = ""
with open("README.md") as rf:
    readme = rf.read()

requirements = ["langfuse==2.*"]

setup(
    name="totvs-dta-utils",
    author="TotvsLabs",
    version="0.1.1",
    author_email="info@totvs.ai",
    python_requires=">=3.10",
    description="Lib for integration with DTA services",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    keywords=[
        "dta",
    ],
    packages=[
        "dta_utils",
        "dta_utils.services",
        "dta_utils.services.assets",
        "dta_utils.services.assets.chat_assets",
        "dta_utils.services.assets.chat_assets.Nunito_Sans",
    ],
    url="https://github.com/totvs-ai/dta-utils-python",
)
